//fast io routines
//i hate to capture this actions

#include "driver.h"

#define VALID_FAST_IO_DISPATCH_HANDLER(_FastIoDispatchPtr, _FieldName) \
    (((_FastIoDispatchPtr) != NULL) && \
     (((_FastIoDispatchPtr)->SizeOfFastIoDispatch) >= \
            (FIELD_OFFSET(FAST_IO_DISPATCH, _FieldName) + sizeof(void *))) && \
     ((_FastIoDispatchPtr)->_FieldName != NULL))

BOOLEAN FastIoCheckIfPossible (
	IN PFILE_OBJECT FileObject,
	IN PLARGE_INTEGER FileOffset,
	IN ULONG Length,
	IN BOOLEAN Wait,
	IN ULONG LockKey,
	IN BOOLEAN CheckForReadOperation,
	OUT PIO_STATUS_BLOCK IoStatus,
	IN PDEVICE_OBJECT DeviceObject
	)
{
	PDEVICE_OBJECT nextDeviceObject;
	PFAST_IO_DISPATCH fastIoDispatch;
	PAGED_CODE();
		ENTER_FUNCTION();
	if (DeviceObject->DeviceExtension) {
		ASSERT(IS_MY_DEVICE_OBJECT( DeviceObject ));

		//
		//  Pass through logic for this type of Fast I/O
		//

		nextDeviceObject = ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject;
		ASSERT(nextDeviceObject);

		fastIoDispatch = nextDeviceObject->DriverObject->FastIoDispatch;

		if (VALID_FAST_IO_DISPATCH_HANDLER( fastIoDispatch, FastIoCheckIfPossible )) {

			return (fastIoDispatch->FastIoCheckIfPossible)(
						FileObject,
						FileOffset,
						Length,
						Wait,
						LockKey,
						CheckForReadOperation,
						IoStatus,
						nextDeviceObject );
		}
	}
	return FALSE;
}


BOOLEAN
FastIoRead (
	IN PFILE_OBJECT FileObject,
	IN PLARGE_INTEGER FileOffset,
	IN ULONG Length,
	IN BOOLEAN Wait,
	IN ULONG LockKey,
	OUT PVOID Buffer,
	OUT PIO_STATUS_BLOCK IoStatus,
	IN PDEVICE_OBJECT DeviceObject
	)
{
	PDEVICE_OBJECT nextDeviceObject;
	PFAST_IO_DISPATCH fastIoDispatch;

	PAGED_CODE();
	
		ENTER_FUNCTION();

	if (DeviceObject->DeviceExtension) {

		ASSERT(IS_MY_DEVICE_OBJECT( DeviceObject ));

		//
		//  Pass through logic for this type of Fast I/O
		//

		nextDeviceObject = ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject;
		ASSERT(nextDeviceObject);

		fastIoDispatch = nextDeviceObject->DriverObject->FastIoDispatch;

		if (VALID_FAST_IO_DISPATCH_HANDLER( fastIoDispatch, FastIoRead )) {

			return (fastIoDispatch->FastIoRead)(
						FileObject,
						FileOffset,
						Length,
						Wait,
						LockKey,
						Buffer,
						IoStatus,
						nextDeviceObject );
		}
	}
	return FALSE;
}


BOOLEAN
FastIoWrite (
	IN PFILE_OBJECT FileObject,
	IN PLARGE_INTEGER FileOffset,
	IN ULONG Length,
	IN BOOLEAN Wait,
	IN ULONG LockKey,
	IN PVOID Buffer,
	OUT PIO_STATUS_BLOCK IoStatus,
	IN PDEVICE_OBJECT DeviceObject
	)
{
	PDEVICE_OBJECT nextDeviceObject;
	PFAST_IO_DISPATCH fastIoDispatch;

	PAGED_CODE();
		ENTER_FUNCTION();

	if (DeviceObject->DeviceExtension) {

		ASSERT(IS_MY_DEVICE_OBJECT( DeviceObject ));

		//
		//  Pass through logic for this type of Fast I/O
		//

		nextDeviceObject = ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject;
		ASSERT(nextDeviceObject);

		fastIoDispatch = nextDeviceObject->DriverObject->FastIoDispatch;

		if (VALID_FAST_IO_DISPATCH_HANDLER( fastIoDispatch, FastIoWrite )) {

			return (fastIoDispatch->FastIoWrite)(
						FileObject,
						FileOffset,
						Length,
						Wait,
						LockKey,
						Buffer,
						IoStatus,
						nextDeviceObject );
		}
	}
	return FALSE;
}


BOOLEAN
FastIoQueryBasicInfo (
	IN PFILE_OBJECT FileObject,
	IN BOOLEAN Wait,
	OUT PFILE_BASIC_INFORMATION Buffer,
	OUT PIO_STATUS_BLOCK IoStatus,
	IN PDEVICE_OBJECT DeviceObject
	)
{
	PDEVICE_OBJECT nextDeviceObject;
	PFAST_IO_DISPATCH fastIoDispatch;

	PAGED_CODE();
		ENTER_FUNCTION();

	if (DeviceObject->DeviceExtension) {

		ASSERT(IS_MY_DEVICE_OBJECT( DeviceObject ));

		//
		//  Pass through logic for this type of Fast I/O
		//

		nextDeviceObject = ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject;
		ASSERT(nextDeviceObject);

		fastIoDispatch = nextDeviceObject->DriverObject->FastIoDispatch;

		if (VALID_FAST_IO_DISPATCH_HANDLER( fastIoDispatch, FastIoQueryBasicInfo )) {

			return (fastIoDispatch->FastIoQueryBasicInfo)(
						FileObject,
						Wait,
						Buffer,
						IoStatus,
						nextDeviceObject );
		}
	}
	return FALSE;
}


BOOLEAN
FastIoQueryStandardInfo (
	IN PFILE_OBJECT FileObject,
	IN BOOLEAN Wait,
	OUT PFILE_STANDARD_INFORMATION Buffer,
	OUT PIO_STATUS_BLOCK IoStatus,
	IN PDEVICE_OBJECT DeviceObject
	)
{
	PDEVICE_OBJECT nextDeviceObject;
	PFAST_IO_DISPATCH fastIoDispatch;

	PAGED_CODE();
		ENTER_FUNCTION();

	if (DeviceObject->DeviceExtension) {

		ASSERT(IS_MY_DEVICE_OBJECT( DeviceObject ));

		//
		//  Pass through logic for this type of Fast I/O
		//

		nextDeviceObject = ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject;
		ASSERT(nextDeviceObject);

		fastIoDispatch = nextDeviceObject->DriverObject->FastIoDispatch;

		if (VALID_FAST_IO_DISPATCH_HANDLER( fastIoDispatch,
											FastIoQueryStandardInfo )) {

			return (fastIoDispatch->FastIoQueryStandardInfo)(
						FileObject,
						Wait,
						Buffer,
						IoStatus,
						nextDeviceObject );
		}
	}
	return FALSE;
}


BOOLEAN
FastIoLock (
	IN PFILE_OBJECT FileObject,
	IN PLARGE_INTEGER FileOffset,
	IN PLARGE_INTEGER Length,
	PEPROCESS ProcessId,
	ULONG Key,
	BOOLEAN FailImmediately,
	BOOLEAN ExclusiveLock,
	OUT PIO_STATUS_BLOCK IoStatus,
	IN PDEVICE_OBJECT DeviceObject
	)
{
	PDEVICE_OBJECT nextDeviceObject;
	PFAST_IO_DISPATCH fastIoDispatch;

	PAGED_CODE();
		ENTER_FUNCTION();

	if (DeviceObject->DeviceExtension) {

		ASSERT(IS_MY_DEVICE_OBJECT( DeviceObject ));

		//
		//  Pass through logic for this type of Fast I/O
		//

		nextDeviceObject = ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject;
		ASSERT(nextDeviceObject);

		fastIoDispatch = nextDeviceObject->DriverObject->FastIoDispatch;

		if (VALID_FAST_IO_DISPATCH_HANDLER( fastIoDispatch, FastIoLock )) {

			return (fastIoDispatch->FastIoLock)(
						FileObject,
						FileOffset,
						Length,
						ProcessId,
						Key,
						FailImmediately,
						ExclusiveLock,
						IoStatus,
						nextDeviceObject );
		}
	}
	return FALSE;
}


BOOLEAN
FastIoUnlockSingle (
	IN PFILE_OBJECT FileObject,
	IN PLARGE_INTEGER FileOffset,
	IN PLARGE_INTEGER Length,
	PEPROCESS ProcessId,
	ULONG Key,
	OUT PIO_STATUS_BLOCK IoStatus,
	IN PDEVICE_OBJECT DeviceObject
	)
{
	PDEVICE_OBJECT nextDeviceObject;
	PFAST_IO_DISPATCH fastIoDispatch;

	PAGED_CODE();
		ENTER_FUNCTION();

	if (DeviceObject->DeviceExtension) {

		ASSERT(IS_MY_DEVICE_OBJECT( DeviceObject ));

		//
		//  Pass through logic for this type of Fast I/O
		//

		nextDeviceObject = ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject;
		ASSERT(nextDeviceObject);

		fastIoDispatch = nextDeviceObject->DriverObject->FastIoDispatch;

		if (VALID_FAST_IO_DISPATCH_HANDLER( fastIoDispatch,
											FastIoUnlockSingle )) {

			return (fastIoDispatch->FastIoUnlockSingle)(
						FileObject,
						FileOffset,
						Length,
						ProcessId,
						Key,
						IoStatus,
						nextDeviceObject );
		}
	}
	return FALSE;
}


BOOLEAN
FastIoUnlockAll (
	IN PFILE_OBJECT FileObject,
	PEPROCESS ProcessId,
	OUT PIO_STATUS_BLOCK IoStatus,
	IN PDEVICE_OBJECT DeviceObject
	)
{
	PDEVICE_OBJECT nextDeviceObject;
	PFAST_IO_DISPATCH fastIoDispatch;

	PAGED_CODE();
		ENTER_FUNCTION();

	if (DeviceObject->DeviceExtension) {

		ASSERT(IS_MY_DEVICE_OBJECT( DeviceObject ));

		//
		//  Pass through logic for this type of Fast I/O
		//

		nextDeviceObject = ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject;

		if (nextDeviceObject) {

			fastIoDispatch = nextDeviceObject->DriverObject->FastIoDispatch;

			if (VALID_FAST_IO_DISPATCH_HANDLER( fastIoDispatch,
												FastIoUnlockAll )) {

				return (fastIoDispatch->FastIoUnlockAll)(
							FileObject,
							ProcessId,
							IoStatus,
							nextDeviceObject );
			}
		}
	}
	return FALSE;
}


BOOLEAN
FastIoUnlockAllByKey (
	IN PFILE_OBJECT FileObject,
	PVOID ProcessId,
	ULONG Key,
	OUT PIO_STATUS_BLOCK IoStatus,
	IN PDEVICE_OBJECT DeviceObject
	)
{
	PDEVICE_OBJECT nextDeviceObject;
	PFAST_IO_DISPATCH fastIoDispatch;

	PAGED_CODE();
		ENTER_FUNCTION();

	if (DeviceObject->DeviceExtension) {

		ASSERT(IS_MY_DEVICE_OBJECT( DeviceObject ));

		//
		//  Pass through logic for this type of Fast I/O
		//

		nextDeviceObject = ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject;
		ASSERT(nextDeviceObject);

		fastIoDispatch = nextDeviceObject->DriverObject->FastIoDispatch;

		if (VALID_FAST_IO_DISPATCH_HANDLER( fastIoDispatch,
											FastIoUnlockAllByKey )) {

			return (fastIoDispatch->FastIoUnlockAllByKey)(
						FileObject,
						ProcessId,
						Key,
						IoStatus,
						nextDeviceObject );
		}
	}
	return FALSE;
}


BOOLEAN
FastIoDeviceControl (
	IN PFILE_OBJECT FileObject,
	IN BOOLEAN Wait,
	IN PVOID InputBuffer OPTIONAL,
	IN ULONG InputBufferLength,
	OUT PVOID OutputBuffer OPTIONAL,
	IN ULONG OutputBufferLength,
	IN ULONG IoControlCode,
	OUT PIO_STATUS_BLOCK IoStatus,
	IN PDEVICE_OBJECT DeviceObject
	)
{
	PDEVICE_OBJECT nextDeviceObject;
	PFAST_IO_DISPATCH fastIoDispatch;

	PAGED_CODE();
		ENTER_FUNCTION();

	if (DeviceObject->DeviceExtension) {

		ASSERT(IS_MY_DEVICE_OBJECT( DeviceObject ));

		//
		//  Pass through logic for this type of Fast I/O
		//

		nextDeviceObject = ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject;
		ASSERT(nextDeviceObject);

		fastIoDispatch = nextDeviceObject->DriverObject->FastIoDispatch;

		if (VALID_FAST_IO_DISPATCH_HANDLER( fastIoDispatch,
											FastIoDeviceControl )) {

			return (fastIoDispatch->FastIoDeviceControl)(
						FileObject,
						Wait,
						InputBuffer,
						InputBufferLength,
						OutputBuffer,
						OutputBufferLength,
						IoControlCode,
						IoStatus,
						nextDeviceObject );
		}
	}
	return FALSE;
}


VOID
FastIoDetachDevice (
	IN PDEVICE_OBJECT SourceDevice,
	IN PDEVICE_OBJECT TargetDevice
	)
{
	PDEVICE_EXTENSION devExt;

	PAGED_CODE();
		ENTER_FUNCTION();
		
		//not supported.
		return ;

		/*
	ASSERT(IS_MY_DEVICE_OBJECT( SourceDevice ));

	devExt = SourceDevice->DeviceExtension;

	//
	//  Display name information
	//

	DbgPrint( "FastIoDetachDevice: Detaching %p from volume   %p \"%wZ\"\n",
				   SourceDevice,
				   TargetDevice,
				   &devExt->NLExtHeader.DeviceName );

	//
	//  Detach from the file system's volume device object.
	//

	//CleanupMountedDevice( SourceDevice );
	IoDetachDevice( TargetDevice );
	IoDeleteDevice( SourceDevice );
	*/
}


BOOLEAN
FastIoQueryNetworkOpenInfo (
	IN PFILE_OBJECT FileObject,
	IN BOOLEAN Wait,
	OUT PFILE_NETWORK_OPEN_INFORMATION Buffer,
	OUT PIO_STATUS_BLOCK IoStatus,
	IN PDEVICE_OBJECT DeviceObject
	)
{
	PDEVICE_OBJECT nextDeviceObject;
	PFAST_IO_DISPATCH fastIoDispatch;

	PAGED_CODE();
		ENTER_FUNCTION();

	if (DeviceObject->DeviceExtension) {

		ASSERT(IS_MY_DEVICE_OBJECT( DeviceObject ));

		//
		//  Pass through logic for this type of Fast I/O
		//

		nextDeviceObject = ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject;
		ASSERT(nextDeviceObject);

		fastIoDispatch = nextDeviceObject->DriverObject->FastIoDispatch;

		if (VALID_FAST_IO_DISPATCH_HANDLER( fastIoDispatch,
											FastIoQueryNetworkOpenInfo )) {

			return (fastIoDispatch->FastIoQueryNetworkOpenInfo)(
						FileObject,
						Wait,
						Buffer,
						IoStatus,
						nextDeviceObject );
		}
	}
	return FALSE;
}


BOOLEAN
FastIoMdlRead (
	IN PFILE_OBJECT FileObject,
	IN PLARGE_INTEGER FileOffset,
	IN ULONG Length,
	IN ULONG LockKey,
	OUT PMDL *MdlChain,
	OUT PIO_STATUS_BLOCK IoStatus,
	IN PDEVICE_OBJECT DeviceObject
	)
{
	PDEVICE_OBJECT nextDeviceObject;
	PFAST_IO_DISPATCH fastIoDispatch;

	PAGED_CODE();
		ENTER_FUNCTION();

	if (DeviceObject->DeviceExtension) {

		ASSERT(IS_MY_DEVICE_OBJECT( DeviceObject ));

		//
		//  Pass through logic for this type of Fast I/O
		//

		nextDeviceObject = ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject;
		ASSERT(nextDeviceObject);

		fastIoDispatch = nextDeviceObject->DriverObject->FastIoDispatch;

		if (VALID_FAST_IO_DISPATCH_HANDLER( fastIoDispatch, MdlRead )) {

			return (fastIoDispatch->MdlRead)(
						FileObject,
						FileOffset,
						Length,
						LockKey,
						MdlChain,
						IoStatus,
						nextDeviceObject );
		}
	}
	return FALSE;
}


BOOLEAN
FastIoMdlReadComplete (
	IN PFILE_OBJECT FileObject,
	IN PMDL MdlChain,
	IN PDEVICE_OBJECT DeviceObject
	)
{
	PDEVICE_OBJECT nextDeviceObject;
	PFAST_IO_DISPATCH fastIoDispatch;
	
		ENTER_FUNCTION();

	if (DeviceObject->DeviceExtension) {

		ASSERT(IS_MY_DEVICE_OBJECT( DeviceObject ));

		//
		//  Pass through logic for this type of Fast I/O
		//

		nextDeviceObject = ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject;
		ASSERT(nextDeviceObject);

		fastIoDispatch = nextDeviceObject->DriverObject->FastIoDispatch;

		if (VALID_FAST_IO_DISPATCH_HANDLER( fastIoDispatch, MdlReadComplete )) {

			return (fastIoDispatch->MdlReadComplete)(
						FileObject,
						MdlChain,
						nextDeviceObject );
		}
	}
	return FALSE;
}


BOOLEAN
FastIoPrepareMdlWrite (
	IN PFILE_OBJECT FileObject,
	IN PLARGE_INTEGER FileOffset,
	IN ULONG Length,
	IN ULONG LockKey,
	OUT PMDL *MdlChain,
	OUT PIO_STATUS_BLOCK IoStatus,
	IN PDEVICE_OBJECT DeviceObject
	)
{
	PDEVICE_OBJECT nextDeviceObject;
	PFAST_IO_DISPATCH fastIoDispatch;

	PAGED_CODE();
		ENTER_FUNCTION();

	if (DeviceObject->DeviceExtension) {

		ASSERT(IS_MY_DEVICE_OBJECT( DeviceObject ));

		//
		//  Pass through logic for this type of Fast I/O
		//

		nextDeviceObject = ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject;
		ASSERT(nextDeviceObject);

		fastIoDispatch = nextDeviceObject->DriverObject->FastIoDispatch;

		if (VALID_FAST_IO_DISPATCH_HANDLER( fastIoDispatch, PrepareMdlWrite )) {

			return (fastIoDispatch->PrepareMdlWrite)(
						FileObject,
						FileOffset,
						Length,
						LockKey,
						MdlChain,
						IoStatus,
						nextDeviceObject );
		}
	}
	return FALSE;
}


BOOLEAN
FastIoMdlWriteComplete (
	IN PFILE_OBJECT FileObject,
	IN PLARGE_INTEGER FileOffset,
	IN PMDL MdlChain,
	IN PDEVICE_OBJECT DeviceObject
	)
{
	PDEVICE_OBJECT nextDeviceObject;
	PFAST_IO_DISPATCH fastIoDispatch;

	PAGED_CODE();
		ENTER_FUNCTION();

	if (DeviceObject->DeviceExtension) {

		ASSERT(IS_MY_DEVICE_OBJECT( DeviceObject ));

		//
		//  Pass through logic for this type of Fast I/O
		//

		nextDeviceObject = ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject;
		ASSERT(nextDeviceObject);

		fastIoDispatch = nextDeviceObject->DriverObject->FastIoDispatch;

		if (VALID_FAST_IO_DISPATCH_HANDLER( fastIoDispatch,
											MdlWriteComplete )) {

			return (fastIoDispatch->MdlWriteComplete)(
						FileObject,
						FileOffset,
						MdlChain,
						nextDeviceObject );
		}
	}
	return FALSE;
}


/*******************************************************************************
		UNIMPLEMENTED FAST IO ROUTINES

		The following four Fast IO routines are for compression on the wire
		which is not yet implemented in NT.

		NOTE:  It is highly recommended that you include these routines (which
			   do a pass-through call) so your filter will not need to be
			   modified in the future when this functionality is implemented in
			   the OS.

		FastIoReadCompressed, FastIoWriteCompressed,
		FastIoMdlReadCompleteCompressed, FastIoMdlWriteCompleteCompressed
*******************************************************************************/


BOOLEAN
FastIoReadCompressed (
	IN PFILE_OBJECT FileObject,
	IN PLARGE_INTEGER FileOffset,
	IN ULONG Length,
	IN ULONG LockKey,
	OUT PVOID Buffer,
	OUT PMDL *MdlChain,
	OUT PIO_STATUS_BLOCK IoStatus,
	OUT struct _COMPRESSED_DATA_INFO *CompressedDataInfo,
	IN ULONG CompressedDataInfoLength,
	IN PDEVICE_OBJECT DeviceObject
	)
{
	PDEVICE_OBJECT nextDeviceObject;
	PFAST_IO_DISPATCH fastIoDispatch;

	PAGED_CODE();
		ENTER_FUNCTION();

	if (DeviceObject->DeviceExtension) {

		ASSERT(IS_MY_DEVICE_OBJECT( DeviceObject ));

		//
		//  Pass through logic for this type of Fast I/O
		//

		nextDeviceObject = ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject;
		ASSERT(nextDeviceObject);

		fastIoDispatch = nextDeviceObject->DriverObject->FastIoDispatch;

		if (VALID_FAST_IO_DISPATCH_HANDLER( fastIoDispatch,
											FastIoReadCompressed )) {

			return (fastIoDispatch->FastIoReadCompressed)(
						FileObject,
						FileOffset,
						Length,
						LockKey,
						Buffer,
						MdlChain,
						IoStatus,
						CompressedDataInfo,
						CompressedDataInfoLength,
						nextDeviceObject );
		}
	}
	return FALSE;
}


BOOLEAN
FastIoWriteCompressed (
	IN PFILE_OBJECT FileObject,
	IN PLARGE_INTEGER FileOffset,
	IN ULONG Length,
	IN ULONG LockKey,
	IN PVOID Buffer,
	OUT PMDL *MdlChain,
	OUT PIO_STATUS_BLOCK IoStatus,
	IN struct _COMPRESSED_DATA_INFO *CompressedDataInfo,
	IN ULONG CompressedDataInfoLength,
	IN PDEVICE_OBJECT DeviceObject
	)
{
	PDEVICE_OBJECT nextDeviceObject;
	PFAST_IO_DISPATCH fastIoDispatch;

	PAGED_CODE();
		ENTER_FUNCTION();

	if (DeviceObject->DeviceExtension) {

		ASSERT(IS_MY_DEVICE_OBJECT( DeviceObject ));

		//
		//  Pass through logic for this type of Fast I/O
		//

		nextDeviceObject = ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject;
		ASSERT(nextDeviceObject);

		fastIoDispatch = nextDeviceObject->DriverObject->FastIoDispatch;

		if (VALID_FAST_IO_DISPATCH_HANDLER( fastIoDispatch,
											FastIoWriteCompressed )) {

			return (fastIoDispatch->FastIoWriteCompressed)(
						FileObject,
						FileOffset,
						Length,
						LockKey,
						Buffer,
						MdlChain,
						IoStatus,
						CompressedDataInfo,
						CompressedDataInfoLength,
						nextDeviceObject );
		}
	}
	return FALSE;
}


BOOLEAN
FastIoMdlReadCompleteCompressed (
	IN PFILE_OBJECT FileObject,
	IN PMDL MdlChain,
	IN PDEVICE_OBJECT DeviceObject
	)
{
	PDEVICE_OBJECT nextDeviceObject;
	PFAST_IO_DISPATCH fastIoDispatch;
	
		ENTER_FUNCTION();

	if (DeviceObject->DeviceExtension) {

		ASSERT(IS_MY_DEVICE_OBJECT( DeviceObject ));

		//
		//  Pass through logic for this type of Fast I/O
		//

		nextDeviceObject = ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject;
		ASSERT(nextDeviceObject);

		fastIoDispatch = nextDeviceObject->DriverObject->FastIoDispatch;

		if (VALID_FAST_IO_DISPATCH_HANDLER( fastIoDispatch,
											MdlReadCompleteCompressed )) {

			return (fastIoDispatch->MdlReadCompleteCompressed)(
						FileObject,
						MdlChain,
						nextDeviceObject );
		}
	}
	return FALSE;
}


BOOLEAN
FastIoMdlWriteCompleteCompressed (
	IN PFILE_OBJECT FileObject,
	IN PLARGE_INTEGER FileOffset,
	IN PMDL MdlChain,
	IN PDEVICE_OBJECT DeviceObject
	)
{
	PDEVICE_OBJECT nextDeviceObject;
	PFAST_IO_DISPATCH fastIoDispatch;
	
		ENTER_FUNCTION();

	if (DeviceObject->DeviceExtension) {

		ASSERT(IS_MY_DEVICE_OBJECT( DeviceObject ));

		//
		//  Pass through logic for this type of Fast I/O
		//

		nextDeviceObject = ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject;
		ASSERT(nextDeviceObject);

		fastIoDispatch = nextDeviceObject->DriverObject->FastIoDispatch;

		if (VALID_FAST_IO_DISPATCH_HANDLER( fastIoDispatch,
											MdlWriteCompleteCompressed )) {

			return (fastIoDispatch->MdlWriteCompleteCompressed)(
						FileObject,
						FileOffset,
						MdlChain,
						nextDeviceObject );
		}
	}
	return FALSE;
}


BOOLEAN
FastIoQueryOpen (
	IN PIRP Irp,
	OUT PFILE_NETWORK_OPEN_INFORMATION NetworkInformation,
	IN PDEVICE_OBJECT DeviceObject
	)
{
	PDEVICE_OBJECT nextDeviceObject;
	PFAST_IO_DISPATCH fastIoDispatch;
	BOOLEAN result;

	PAGED_CODE();
		ENTER_FUNCTION();

	if (DeviceObject->DeviceExtension) {

		ASSERT(IS_MY_DEVICE_OBJECT( DeviceObject ));

		//
		//  Pass through logic for this type of Fast I/O
		//

		nextDeviceObject = ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject;
		ASSERT(nextDeviceObject);

		fastIoDispatch = nextDeviceObject->DriverObject->FastIoDispatch;

		if (VALID_FAST_IO_DISPATCH_HANDLER( fastIoDispatch, FastIoQueryOpen )) {

			PIO_STACK_LOCATION irpSp = IoGetCurrentIrpStackLocation( Irp );

			//
			//  Before calling the next filter, we must make sure their device
			//  object is in the current stack entry for the given IRP
			//

			irpSp->DeviceObject = nextDeviceObject;

			result = (fastIoDispatch->FastIoQueryOpen)(
						Irp,
						NetworkInformation,
						nextDeviceObject );

			//
			//  Always restore the IRP back to our device object
			//

			irpSp->DeviceObject = DeviceObject;
			return result;
		}
	}
	return FALSE;
}

FAST_IO_DISPATCH fast_io = {
	sizeof(FAST_IO_DISPATCH), 
	FastIoCheckIfPossible,
	FastIoRead,
	FastIoWrite,
	FastIoQueryBasicInfo,
	FastIoQueryStandardInfo,
	FastIoLock,
	FastIoUnlockSingle,
	FastIoUnlockAll,
	FastIoUnlockAllByKey,
	FastIoDeviceControl,
	NULL, //FastIoAcquireFile,
	NULL, //FastIoReleaseFile,
	FastIoDetachDevice,
	// new for NT 4.0
	FastIoQueryNetworkOpenInfo,
	NULL, //FastIoAcquireForModWrite,
	FastIoMdlRead,
	FastIoMdlReadComplete,
	FastIoPrepareMdlWrite,
	FastIoMdlWriteComplete,
	FastIoReadCompressed,
	FastIoWriteCompressed,
	FastIoMdlReadCompleteCompressed,
	FastIoMdlWriteCompleteCompressed,
	FastIoQueryOpen,
	NULL, //FastIoReleaseForModWrite,
	NULL, //FastIoAcquireForCcFlush,
	NULL, //FastIoReleaseForCcFlush
};
