/*
 *  list.c
 *
 *  Do with array list.
 *
 *  Copyright (C) 2008  Huang Guan
 *
 *  2008-02-5 Created.
 *  Note: Don't use these functions while another is using it..
 *        No any mutual symbols support.
 *
 */
#include "driver.h"
#include "rlist.h"
#include "util.h"


//
int rlist_init( rlist_t* l, int size, remove_item remover, compare_item comparer )
{
	if( l == NULL ) return -1;
	l->list = (int*) kmem_get( size * sizeof(int) );
	l->size = size;
	l->head = 0;
	l->tail = 0;
	l->remover = remover;
	l->comparer = comparer;
	return 0;
}

//
#define NEXT ( (l->tail+1)%l->size )
int rlist_append( rlist_t* l, int v )
{
	if( l == NULL ) return -1;
	//set value
	l->list[ l->tail ] = v;
	//move
	if( NEXT == l->head ) //full?
	{
		if( l->remover )
			l->remover( l->list[ l->head ] );
		l->head = (l->head + 1)%l->size;
	}
	l->tail = NEXT;
	return 0;
}

//
void rlist_cleanup( rlist_t* l )
{
	int i;
	if( l == NULL ) return ;
	for( i=l->head; i!=l->tail; i=(i+1)%l->size )
	{
		if( l->remover )
			l->remover( l->list[i] );
	}
	kmem_free( l->list );
	l->list = NULL;
}

//
int rlist_find( rlist_t * l, int v )
{
	int i;
	if( l == NULL ) return -1;
	for( i=l->head; i!=l->tail; i=(i+1)%l->size )
	{
		if( l->comparer &&
			l->comparer( v, l->list[i] )==0 )
		{
			return 1;
		}else{
			if( l->list[i] == v )
				return 1;
		}
	}
	return 0;
}

