#include "driver.h"
#include "user.h"
#include "process.h"
#include "util.h"

#define MAX_PATH_LEN 1024
#define MAX_OPER_ENTRY 50

enum{
	WS_NONE = 0,
	WS_UData,
	WS_KAsk,
	WS_UResult,
	WS_KComp,
	WS_Doing,
};

struct oper_entry{
	wchar_t name[MAX_PATH_LEN];
	wchar_t file[MAX_PATH_LEN];
	int oper;
};
static struct oper_entry* oper_list = NULL;
static int	oper_count;

static int mut_asking;
static state_t worker;

static wchar_t proc_name[MAX_PATH_LEN], oper_file[MAX_PATH_LEN];
static int result, init = 0;
static int proc_master;

int user_init()
{
	mut_init(&mut_asking);
	init = 1;
	state_init( &worker );
	state_set( &worker, 0 );
	proc_master = (int)PsGetCurrentProcess();
	dprint("MasterProc: %ws", process_file_path() );
	oper_list = (struct oper_entry*)kmem_get( sizeof(struct oper_entry)* MAX_OPER_ENTRY );
	oper_count = 0;
	return 0;
}

int user_add_oper( wchar_t* name, wchar_t* file, int oper )
{
	int i;
	if( !name || !file || !oper_list )
		return 0;
	i = oper_count;
	wcsncpy( oper_list[i].name, name, MAX_PATH_LEN );
	wcsncpy( oper_list[i].file, file, MAX_PATH_LEN );
	oper_list[i].oper = oper;
	oper_count ++;
	return i;
}

int user_find_oper( wchar_t* name, wchar_t* file, int* oper )
{
	int i;
	if( !name || !file || !oper_list )
		return 0;
	for( i=0; i<oper_count; i++ )
	{
		if( (_wcsicmp( name, oper_list[i].name ) == 0) &&
			(_wcsicmp( file, oper_list[i].file ) == 0) )
		{
			*oper = oper_list[i].oper;
			return 1;
		}
	}
	return 0;
}

int user_remove_oper( int i )
{
	if( i<0 || i>=oper_count || !oper_list) return -1;
	if( i!=oper_count-1 )
		oper_list[i] = oper_list[oper_count-1];
	oper_count --;
	return 0;
}

void user_cleanup()
{
	if( oper_list )
		kmem_free( oper_list );
	oper_list = NULL;
}

int user_ask( wchar_t* name, wchar_t* file )
{
	int ret;
	if( name == NULL || file == NULL )
		return 0;
	dprint("proc ask: %ws for %ws", name, file );
	if( !init  )
		return 0;
	if( (int)PsGetCurrentProcess() == proc_master ){
		dprint("Ah!! Master!!");
		return 1; //allow
	}
	//we do with one at one time.
	mut_down( &mut_asking );
	if( user_find_oper( name, file, &result ) )
	{
		dprint("Found previous oper.");
		mut_up( &mut_asking );
		return result;
	}
	//dprint("k wait for state");
	if( state_wait( &worker, WS_UData )<0 ){
		dprint("timeout to wait WS_DATA.");
		mut_up( &mut_asking );
		return 1;
	}
	state_set( &worker, WS_Doing );
	mut_up( &mut_asking );
	//ok, no race condition here!!
	wcsncpy( proc_name, name, MAX_PATH_LEN );
	wcsncpy( oper_file, file, MAX_PATH_LEN );
	//tell the user thread that data is ready
	state_set( &worker, WS_KAsk );
	//wait for reply
	//dprint("[%x] k wait for reply", (int)PsGetCurrentProcess() );
	if( state_wait( &worker, WS_UResult ) < 0 ){
		//timeout
		return 1;
	}
	//dprint( "[%x] k wait ok %d.", (int)PsGetCurrentProcess(), result );
	if( oper_count == MAX_OPER_ENTRY )
	{
		dprint("oper full");
		user_remove_oper( 0 );
	}
	user_add_oper( name, file, result );
	ret = result;
	state_set( &worker, WS_KComp );
	//ok. replied
	return ret;
}

int user_get( char* buf, int size )
{
	if( size<MAX_PATH_LEN*sizeof(wchar_t)*2 )
		return 0;
	if( !init )
		user_init();
	if( worker.state != WS_KAsk )
	{
		if( worker.state != WS_UData )
			state_set( &worker, WS_UData );
		ksleep( 100 );
		if( worker.state != WS_KAsk )
		{
			return 0;
		}
	}
	//wait for data ready
	//
	wcsncpy( (wchar_t*)buf, proc_name, MAX_PATH_LEN );
	wcsncpy( (wchar_t*)(buf+MAX_PATH_LEN*sizeof(wchar_t)), oper_file, MAX_PATH_LEN);
	//ok.
	//dprint("u got data.");
	return size;
}

void user_result( int r )
{
	result = r;
	dprint("u result %d", r );
	state_set( &worker, WS_UResult );
	state_wait( &worker, WS_KComp );
}




