
#define PTHREAD

#include <windows.h>
#include <winioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef PTHREAD
#include <pthread.h>
#endif
#include "../sys/ioctl.h"
#include "../sys/file_node.h"

HANDLE dev;
DWORD err = 0;
int ret;
long bytes;
char version[100];
int get_ver();
int open_dev();
int __declspec(dllexport) __stdcall status(int s);
int close_dev();
int writeout();
int writeproc();
int readin();
#ifdef PTHREAD
pthread_t thread_get;
int thread_run = 0; //1 or 0
int __declspec(dllexport) __stdcall quit_thread();
#endif
int get( wchar_t* name, wchar_t* file );
int do_result(int r);
extern void out_file_node(char* file);


#define MAX_PATH_LEN 1024
#define MAX_SYS_PATH 32
struct sysproc_list{
	int count;
	wchar_t list[MAX_SYS_PATH][MAX_PATH_LEN];
};

static struct sysproc_list * sp_list = NULL;

int __declspec(dllexport) __stdcall do_init(){
	if( open_dev()<0 )
		return -1;
	get_ver();
	printf("Version: %s\n", version);
	file_node_init();
	sp_list = (struct sysproc_list*)malloc( sizeof(struct sysproc_list));
	if( sp_list == NULL )
	{
		printf("Allocate memory failed.");
		return -1;
	}
	sp_list->count = 0;
	return 0;
}

int __declspec(dllexport) __stdcall do_cleanup(){
#ifdef PTHREAD
	quit_thread();
#endif
	file_node_cleanup();
	close_dev();
	free( sp_list );
	sp_list = NULL;
	return 0;
}

int w2b(wchar_t *pw, char *pb, int outlen)
{
	int nRet;
	nRet = WideCharToMultiByte(CP_ACP, WC_COMPOSITECHECK, (LPWSTR)pw, -1, pb, outlen, NULL, NULL);
	if(nRet==0)
	{
		return -1;
	}
	return 0;
}

#ifdef PTHREAD
char g_proc[512], g_file[512], hasMsg = 0, hasResult;
static void get_loop( void* data )
{
	wchar_t wproc[1024], wfile[1024], result=0;
	while( thread_run ){
		if( get(wproc, wfile)<0 )
		{
			//printf("get error.\n");
			continue;
		}
		if( w2b( wproc, g_proc, 512 ) < 0 ){
			printf("error in w2b()\n");
			continue;
		}
		if( w2b( wfile, g_file, 512 ) < 0 ){
			printf("error in w2b()\n");
			continue;
		}
		hasMsg = 1;
		while( hasMsg ){
			if( !thread_run ){
				break;
			}
			sleep( 100 );
		}
		if( !thread_run )
			break;
		do_result( hasResult );
	}
	thread_run = 0;
}

int __declspec(dllexport) __stdcall make_thread(){
	int ret;
	if( !thread_run )
	{
		thread_run = 1;
		hasMsg = 0;
		ret = pthread_create( &thread_get, NULL, (void*)get_loop, NULL );
		if( ret != 0 )
		{
			printf("Create thread ERROR.");
			return -1;
		}
	}
	return 0;
}

int __declspec(dllexport) __stdcall quit_thread()
{
	if( thread_run )
	{
		thread_run = 0;
		status( 1 );	//stop
		sleep( 200 );
	}
	return 0;
}

int __declspec(dllexport) __stdcall poll_msg( int* p, int* f )
{
	if( hasMsg ){
		*p = (int)g_proc;
		*f = (int)g_file;
		return 1;
	}
	if( !thread_run ){
		return -1;
	}
	return 0;
}

int __declspec(dllexport) __stdcall put_result( int r )
{
	hasResult = r;
	hasMsg = 0;
	return 0;
}
#endif

int __declspec(dllexport) __stdcall fn_output( char* str )
{
	readin();
	out_file_node( str );
	return 0;
}

int __declspec(dllexport) __stdcall fn_clean()
{
	file_node_cleanup();
	file_node_init();
	return 0;
}

int __declspec(dllexport) __stdcall fn_write()
{
	return writeout();
}

int __declspec(dllexport) __stdcall fn_add( char* str, int range, int flag )
{
	wchar_t wpath[2048];
	int len;
	len = MultiByteToWideChar(CP_ACP, 0, str, strlen(str), wpath,sizeof(wpath));
	wpath[len] = 0;
	add_prot_node( wpath, (unsigned char)range, (unsigned char)flag );
	return 0;
}

int process_add_sys( wchar_t* p )
{
	int i;
	if( !sp_list || !p ) {
		return -1;
	}
	i = sp_list->count;
	if( i>= MAX_SYS_PATH ) {
		return -2;
	}
	wcsncpy( sp_list->list[i], p, MAX_PATH_LEN );
	sp_list->count ++;
	return 0;
}

int __declspec(dllexport) __stdcall sysproc_add( char* str )
{
	wchar_t wpath[2048];
	int len;
	len = MultiByteToWideChar(CP_ACP, 0, str, strlen(str), wpath,sizeof(wpath));
	wpath[len] = 0;
	return process_add_sys( wpath );
}

int __declspec(dllexport) __stdcall sysproc_clean()
{
	sp_list->count = 0;
	return 0;
}

int __declspec(dllexport) __stdcall sysproc_write()
{
	return writeproc();
}

int __cdecl main(ULONG argc, PCHAR argv[] )
{
	do_init();
	if( argc>1 )
	{
		char* cmd = argv[1];
		char* arg = argv[2];
		if( strcmp( cmd, "status") == 0 )
		{
			int i;
			i = atoi( arg );
			status( i );
		}else if( strcmp( cmd, "input") == 0 )
		{
			FILE *fp = fopen( arg, "r" );
			wchar_t wpath[2048];
			char path[2048];
			int prot, exec;
			int len;
			while( !feof(fp) )
			{
				fscanf( fp, "%s %d %d", path, &prot, &exec );
				len = MultiByteToWideChar(CP_ACP, 0, path, strlen(path), wpath,sizeof(wpath));
				wpath[len] = 0;
				add_prot_node( wpath, (unsigned char)prot, (unsigned char)exec );
			}
			fclose(fp);
			writeout();
			
		}else if( strcmp( cmd, "output") == 0 )
		{
			readin();
			out_file_node( arg );
		}else if( strcmp( cmd, "get") == 0 )
		{
			wchar_t wproc[1024], wfile[1024], result=0;
			char proc[512], file[512];
			while( result!=3 ){
				if( get(wproc, wfile)<0 )
				{
					printf("get error.\n");
					exit(1);
				}
				if( w2b( wproc, proc, 512 ) < 0 ){
					printf("error in w2b()\n");
					exit(1);
				}
				if( w2b( wfile, file, 512 ) < 0 ){
					printf("error in w2b()\n");
					exit(1);
				}
				printf("Process %s", proc );
				printf("\n is trying to write into %s", file );
				printf("\nDo you agree? (1 or 0) ");
				scanf("%d", &result );
				do_result( result );
			}
		}
	}
	do_cleanup();
}

int close_dev()
{
	CloseHandle ( dev );
	return 0;
}

int open_dev()
{	
	// open the device
	dev = CreateFile( "\\\\.\\fsec",
					  GENERIC_READ | GENERIC_WRITE,
					  0,
					  NULL,
					  CREATE_ALWAYS,
					  FILE_ATTRIBUTE_NORMAL,
					  NULL);
	
	if ( dev == INVALID_HANDLE_VALUE ){
		printf ( "Error: CreatFile Failed : %d\n", GetLastError());
		return -1;
	}
	return 0;
}

int get_ver()
{	
	//version
	memset(version, 0, sizeof(version));
	ret = DeviceIoControl ( dev,
						  (DWORD) IOCTL_FSEC_VERSION,
						  &version,
						  sizeof( version),
						  &version,
						  sizeof( version),
						  &bytes, NULL
						  );
	if ( !ret )
	{
		printf ( "Error in DeviceIoControl : %d", GetLastError());
		return -1;
	}
	return 0;
}

int __declspec(dllexport) __stdcall status(int s)
{
	ret = DeviceIoControl ( dev,
						  (DWORD) IOCTL_FSEC_STATUS,
						  &s,
						  sizeof(s),
						  &s,
						  sizeof(s),
						  &bytes, NULL
						  );
	printf("Now status: %d\n", s );
	if ( !ret )
	{
		printf ( "Error in DeviceIoControl : %d", GetLastError());
		return -1;
	}
	return s;
}

struct file_node{
	//link table
	short next;	//next id of the file_node.
	short pre;		//previous id of the file_node.
	short parent;	//parent id of the file_node.
	short child;	//child id of the file_node.
	//internal information
	wchar_t file_name[MAX_FILENAME_LEN];
	unsigned char prot_type;	//indicate the type of protection. ( PT_RDONLY )
	unsigned char prot_exec;   //means only protect executable files
	unsigned char node_type;   //type of node. NT_NORMAL or NT_CUSTOMED
	short  count;	//how many users using this file, the normal node will be removed when count==0.
};

int readin()
{
	int cb;
	char* buf;
	cb = sizeof( struct file_node ) * MAX_FILENODE;
	buf = malloc( cb );
	ret = DeviceIoControl ( dev,
						  (DWORD) IOCTL_FSEC_GETNODES,
						  buf,
						  cb,
						  buf,
						  cb,
						  &bytes, NULL
						  );
	if ( !ret )
	{
		printf ( "Error in DeviceIoControl : %d", GetLastError());
		free( buf );
		return -1;
	}
	set_file_nodes( buf, bytes );
	free( buf );
	return 0;
}

int writeout()
{
	int cb, len;
	char* buf;
	cb = sizeof( struct file_node ) * MAX_FILENODE;
	buf = malloc( cb );
	len = get_file_nodes( buf, cb );
	ret = DeviceIoControl ( dev,
						  (DWORD) IOCTL_FSEC_SETNODES,
						  buf,
						  len,
						  buf,
						  cb,
						  &bytes, NULL
						  );
	if ( !ret )
	{
		printf ( "Error in DeviceIoControl : %d", GetLastError());
		free( buf );
		return -1;
	}
	free( buf );
	return 0;
}

int writeproc()
{
	int cb, len;
	char* buf;
	cb = sizeof(struct sysproc_list);
	if( sp_list == NULL )
		return -2;
	buf = (char*)sp_list;
	len = cb;
	ret = DeviceIoControl ( dev,
						  (DWORD) IOCTL_FSEC_SETSYSPROC,
						  buf,
						  len,
						  buf,
						  cb,
						  &bytes, NULL
						  );
	if ( !ret )
	{
		printf ( "Error in DeviceIoControl : %d", GetLastError());
		return -1;
	}
	return 0;
}

int get( wchar_t* name, wchar_t* file )
{
	int cb;
	char* buf;
	cb = sizeof( wchar_t ) * 1024 * 2;
	buf = malloc( cb );
	ret = DeviceIoControl ( dev,
						  (DWORD) IOCTL_FSEC_USERGET,
						  buf,
						  cb,
						  buf,
						  cb,
						  &bytes, NULL
						  );
	if ( !ret )
	{
		printf ( "Error in DeviceIoControl : %d", GetLastError());
		free( buf );
		return -1;
	}
	if( bytes == 0 ){
		printf("Get failed. bytes==0\n");
		free( buf );
		return -5;
	}
	wcsncpy( name, (wchar_t*)buf, 1024 );
	wcsncpy( file, (wchar_t*)(buf+1024*sizeof(wchar_t)), 1024);
	free( buf );
	return 0;
}

int do_result( int r )
{
	int cb;
	char* buf;
	cb = 4;
	buf = (char*)&r;
	ret = DeviceIoControl ( dev,
						  (DWORD) IOCTL_FSEC_USERRESULT,
						  buf,
						  cb,
						  buf,
						  cb,
						  &bytes, NULL
						  );
	if ( !ret )
	{
		printf ( "Error in DeviceIoControl : %d", GetLastError());
		free( buf );
		return -1;
	}
	return ret;
}
